from __future__ import print_function, annotations

from os import access, R_OK, environ
from os.path import basename, isfile, expanduser, getsize

import json
import logging
import requests
import tomli
from datetime import datetime
from hashlib import sha256
from lief import parse, Binary, ELF, PE, MachO
from numpy import array, vstack
from pandas import DataFrame
from requests import request, Response, HTTPError
from sklearn.metrics.pairwise import cosine_similarity

__version__ = "1.2.1"

re_conf = {
    "apikey": environ.get("REAI_API_KEY", ""),
    "host": environ.get("REAI_API_HOST", "https://api.reveng.ai"),
}


logger = logging.getLogger("REAIT")


class ReaitError(HTTPError):
    def __init__(self, reason: str, end_point: str = None):
        response = Response()

        response.reason = reason
        response.status_code = 404
        response._content = b'{"success": false, "error": "' + \
            reason.encode() + b'"}'
        response.url = (
            f"{re_conf['host']}/{end_point if end_point[0] != '/' else end_point[1:]}"
            if end_point
            else None
        )

        super().__init__(reason, response=response)


def reveng_req(
    req: request,
    end_point: str,
    data: dict = None,
    ex_headers: dict = None,
    params: dict = None,
    json_data: dict = None,
    timeout: int = 60,
    files: dict = None,
) -> Response:
    """
    Constructs and sends a Request
    :param req: Method for the new Request
    :param end_point: Endpoint to add to the base URL
    :param ex_headers: Extended HTTP headers to add
    :param data: Dictionary, list of tuples, bytes, or file-like object to send in the body
    :param params: Dictionary, list of tuples or bytes to send in the query string for the query string
    :param json_data: A JSON serializable Python object to send in the body
    :param timeout: Number of seconds to stop waiting for a Response
    :param files: Dictionary of files to send to the specified URL
    """
    url = f"{re_conf['host']}/{end_point if end_point[0] != '/' else end_point[1:]}"
    headers = {"Authorization": re_conf["apikey"]}

    if ex_headers:
        headers.update(ex_headers)

    logger.debug(
        "Making %s request %s:\n  - headers: %s\n  - data: %s\n  - json_data: %s\n  - params: %s\n  - files: %s",
        req.__name__.upper(),
        url,
        headers,
        data,
        json_data,
        params,
        files,
    )

    response: Response = req(
        url,
        headers=headers,
        json=json_data,
        data=data,
        params=params,
        timeout=timeout,
        files=files,
    )

    logger.debug(
        "Making %s response %s:\n  - headers: %s\n  - status_code: %d\n  - content: %s",
        req.__name__.upper(),
        url,
        response.headers,
        response.status_code,
        response.text,
    )

    return response


def re_hash_check(bin_id: str) -> bool:
    res: Response = reveng_req(
        requests.get, "v1/search", json_data={"sha_256_hash": bin_id}
    )

    if res.ok:
        return any(
            binary["sha_256_hash"] == bin_id for binary in res.json()["query_results"]
        )
    else:
        logger.warning("Bad Request: %s", res.text)

    return False


# Bin_id is referred to as hash in this program - to maintain usage BID = id of a binary bin_id = hash
# Assumes a file has been passed, correct hash only
# Returns the BID of the binary_id (hash)
def re_bid_search(bin_id: str) -> int:
    res: Response = reveng_req(
        requests.get, "v1/search", json_data={"sha_256_hash": bin_id}
    )

    bid = -1

    if res.ok:
        # Filter the result who matches the SHA-256
        binaries = list(
            filter(
                lambda binary: binary["sha_256_hash"] == bin_id,
                res.json()["query_results"],
            )
        )

        # Check only one record is returned
        if len(binaries) == 1:
            binary = binaries[0]
            bid = binary["binary_id"]

            logger.info(
                "Only one record exists, selecting - ID: %d, Name: %s, Creation: %s, Model: %s, Status: %s",
                bid,
                binary["binary_name"],
                binary["creation"],
                binary["model_name"],
                binary["status"],
            )
        elif len(binaries) > 1:
            binaries.sort(
                key=lambda binary: datetime.fromisoformat(
                    binary["creation"]
                ).timestamp(),
                reverse=True,
            )

            logger.info("%d matches found for hash: %s", len(binaries), bin_id)

            options_dict = {}

            for idx, binary in enumerate(binaries):
                logger.info(
                    "[%d] - ID: %d, Name: %s, Creation: %s, Model: %s, Status: %s",
                    idx,
                    binary["binary_id"],
                    binary["binary_name"],
                    binary["creation"],
                    binary["model_name"],
                    binary["status"],
                )

                options_dict[idx] = binary["binary_id"]

            try:
                user_input = input(
                    "[+] Please enter the option you want to use for this operation:"
                )

                option_number = int(user_input)

                bid = options_dict.get(option_number, -1)

                if bid == -1:
                    logger.warning("Invalid option.")
            except Exception:
                bid = options_dict[0]
                logger.warning("Select the most recent analysis - ID: %d", bid)
        else:
            logger.warning("No matches found for hash: %s", bin_id)
    else:
        logger.warning("Bad Request: %s", res.text)

    res.raise_for_status()
    return bid


def RE_delete(fpath: str, binary_id: int = 0) -> Response:
    """
    Delete analysis results for Binary ID in command
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    end_point = f"v1/analyse/{bid}"

    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}", end_point)

    res: Response = reveng_req(requests.delete, end_point)

    if res.ok:
        logger.info("Securely deleted analysis ID %s - %s.", bid, bin_id)
    elif res.status_code == 404:
        logger.warning("Error analysis not found for ID %s - %s.", bid, bin_id)
    else:
        logger.error(
            "Error deleting binary %s under. Server returned %d.",
            bin_id,
            res.status_code,
        )

    res.raise_for_status()
    return res


def RE_analyse(
    fpath: str,
    model_name: str = None,
    isa_options: str = None,
    platform_options: str = None,
    file_options: str = None,
    dynamic_execution: bool = False,
    command_line_args: str = None,
    binary_scope: str = None,
    tags: list = None,
    priority: int = 0,
    duplicate: bool = False,
    symbols: dict = None,
    debug_fpath: str = None,
    skip_scraping: bool = False,
    skip_capabilities: bool = False,
    skip_sbom: bool = False,
    advanced_analysis: bool = False
) -> Response:
    """
    Start analysis job for binary file
    :param fpath: File path for binary to analyse
    :param model_name: Binary model name
    :param isa_options: Executable ISA
    :param file_options: File options
    :param platform_options: OS platform
    :param dynamic_execution: Enable dynamic execution in sandbox during analysis
    :param command_line_args: Command line arguments to pass when running binary sample in the sandbox
    :param binary_scope: Analysis visibility
    :param tags: Assign tags to an analysis
    :param priority: Priority to processing queue
    :param duplicate: Duplicate an existing binary
    :param symbols: JSON object containing the base address and the list of functions
    :param debug_fpath: File path for debug file
    :param skip_scraping: Disable/Enable auto-tagging of binary sample in relevant APIs
    """
    bin_id = re_binary_id(fpath)
    result = re_hash_check(bin_id)

    end_point = "v1/analyse/"

    if result and duplicate is False:
        logger.error(
            "Error, duplicate analysis for %s. To upload again, use the --duplicate flag.",
            bin_id,
        )
        raise ReaitError(f"Duplicate analysis for hash: {bin_id}", end_point)

    filename = basename(fpath)

    params = {
        "file_name": filename,
        "size_in_bytes": getsize(fpath),
        "sha_256_hash": bin_id,
    }

    if debug_fpath and isfile(debug_fpath) and access(debug_fpath, R_OK):
        try:
            debug = RE_upload(debug_fpath).json()

            if debug["success"]:
                params["debug_hash"] = debug["sha_256_hash"]
        except HTTPError:
            pass

    for p_name in (
        "model_name",
        "isa_options",
        "platform_options",
        "file_options",
        "dynamic_execution",
        "command_line_args",
        "binary_scope",
        "tags",
        "priority",
        "symbols",
        "skip_scraping",
        "skip_capabilities",
        "skip_sbom",
        "advanced_analysis"
    ):
        p_value = locals()[p_name]

        if p_value:
            params[p_name] = p_value

    res: Response = reveng_req(requests.post, end_point, json_data=params)
    if res.ok:
        logger.info(
            "Successfully submitted binary for analysis. %s - %s", fpath, bin_id
        )
    elif res.status_code == 400:
        if "error" in res.json().keys():
            logger.warning("Error analysing %s - %s",
                           fpath, res.json()["error"])

    res.raise_for_status()
    return res


def RE_upload(fpath: str) -> Response:
    """
    Upload binary to Server
    :param fpath: File path for binary to analyse
    """
    bin_id = re_binary_id(fpath)
    result = re_hash_check(bin_id)

    if result:
        logger.info(
            "File %s - %s already uploaded. Skipping upload...", fpath, bin_id)

        res = Response()
        res.status_code = 200
        res.url = f"{re_conf['host']}/v1/upload"
        res._content = (
            (
                '{0}"success": true,'
                '"message": "File already uploaded!",'
                '"sha_256_hash": "{1}"{2}'
            )
            .format("{", bin_id, "}")
            .encode()
        )
    else:
        with open(fpath, "rb") as fd:
            res: Response = reveng_req(
                requests.post, "v1/upload", files={"file": fd})

        if res.ok:
            logger.info(
                "Successfully uploaded binary to your account. %s - %s", fpath, bin_id
            )
        elif res.status_code == 400:
            if "error" in res.json().keys():
                logger.warning("Error uploading %s - %s",
                               fpath, res.json()["error"])
        elif res.status_code == 413:
            logger.warning("File too large. Please upload files under 10MB.")
        elif res.status_code == 500:
            logger.error(
                "Internal Server Error. Please contact support. Skipping upload..."
            )

    res.raise_for_status()
    return res


def RE_embeddings(fpath: str, binary_id: int = 0) -> Response:
    """
    Fetch symbol embeddings
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    end_point = f"v1/embeddings/binary/{bid}"

    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}", end_point)

    res: Response = reveng_req(requests.get, end_point)

    if res.status_code == 400:
        logger.warning(
            "Analysis for %s still in progress. Please check the logs (-l) and try again later.",
            bin_id,
        )

    res.raise_for_status()
    return res


def RE_logs(fpath: str, binary_id: int = 0, console: bool = True) -> Response:
    """
    Get the logs for an analysis associated to Binary ID in command
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    :param console: Show response in console
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    end_point = f"v1/logs/{bid}"

    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}", end_point)

    res: Response = reveng_req(requests.get, end_point)

    if res.ok and console:
        logger.info("Logs found for %s:\n%s", bin_id, res.json()["logs"])
    elif res.status_code == 404:
        logger.warning("Error, logs not found for %s.", bin_id)

    res.raise_for_status()
    return res


def RE_cves(fpath: str, binary_id: int = 0) -> Response:
    """
    Check for known CVEs in Binary
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    end_point = f"cves/{bid}"

    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}", end_point)

    res: Response = reveng_req(requests.get, end_point)

    if res.ok:
        cves = json.loads(res.text)
        logger.info("Checking for known CVEs embedded inside %s", fpath)

        if len(cves) == 0:
            logger.info("0 CVEs found.")
        else:
            logger.warning("Warning CVEs found!\n%s", res.text)
    elif res.status_code == 404:
        logger.warning("Error, binary analysis not found for %s.", bin_id)

    res.raise_for_status()
    return res


def RE_status(fpath: str, binary_id: int = 0, console: bool = False) -> Response:
    """
    Get the status of an ongoing binary analysis
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    end_point = f"v1/analyse/status/{bid}"

    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}", end_point)

    res: Response = reveng_req(requests.get, end_point)

    if res.ok and console:
        logger.info("Binary analysis status: %s", res.json()["status"])
    if res.status_code == 400:
        logger.warning(" Error, status not found for %s.", bin_id)

    res.raise_for_status()
    return res


def RE_compute_distance(embedding: list, embeddings: list, nns: int = 5) -> list:
    """
    Compute the cosine distance between source embedding and embedding from binary
    :param embedding: Embedding vector as python list
    :param embeddings: Symbol embeddings
    :param nns: Number of nearest neighbors
    """
    df = DataFrame(data=embeddings)
    np_embedding = array(embedding).reshape(1, -1)
    source_embeddings = vstack(df["embedding"].values)
    closest = (
        cosine_similarity(source_embeddings, np_embedding)
        .squeeze()
        .argsort()[::-1][:nns]
    )
    distances = cosine_similarity(source_embeddings[closest], np_embedding)

    # match closest embeddings with similarity
    closest_df = df.iloc[closest]

    # create json similarity object
    similarities = list(zip(distances, closest_df.index.tolist()))
    json_sims = [
        {
            "similaritiy": float(d[0]),
            "vaddr": int(df.iloc[v]["vaddr"]),
            "name": str(df.iloc[v]["name"]),
            "size": int(df.iloc[v]["size"]),
        }
        for d, v in similarities
    ]
    return json_sims


def RE_nearest_symbols_batch(
    function_ids: list[int],
    nns: int = 5,
    collections: list[int] = None,
    binaries: list[int] = None,
    distance: float = 0.1,
    debug_enabled: bool = False,
) -> Response:
    """
    Get nearest functions to a passed function ids
    :param function_ids: List of function ids
    :param nns: Number of nearest neighbors
    :param collections: List of collections RevEng.AI collection names to search through
    :param distance: How close we want the ANN search to filter for
    :param debug_enabled: ANN Symbol Search, only perform ANN on debug symbols if set
    """
    params = {
        "function_id_list": function_ids,
        "result_per_function": nns,
        "debug_mode": debug_enabled,
        "distance": distance,
    }

    if collections:
        params["collection_search_list"] = collections

    if binaries:
        params["binaries_search_list"] = binaries

    res: Response = reveng_req(
        requests.post, "v1/ann/symbol/batch", json_data=params)

    res.raise_for_status()
    return res


def RE_nearest_functions(
    fpath: str,
    binary_id: int = 0,
    nns: int = 5,
    distance: float = 0.1,
    debug_enabled: bool = False,
) -> Response:
    """
    Get the nearest functions
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    :param nns: Number of nearest neighbors
    :param distance: How close we want the ANN search to filter for
    :param debug_enabled: ANN Symbol Search, only perform ANN on debug symbols if set
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    end_point = f"v1/ann/symbol/{bid}"

    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}", end_point)

    params = {
        "result_per_function": nns,
        "debug_mode": debug_enabled,
        "distance": distance,
    }

    res: Response = reveng_req(requests.post, end_point, json_data=params)

    res.raise_for_status()
    return res


def RE_analyze_functions(fpath: str, binary_id: int = 0) -> Response:
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    end_point = f"v1/analyse/functions/{bid}"

    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}", end_point)

    res: Response = reveng_req(requests.get, end_point)

    res.raise_for_status()
    return res


def RE_SBOM(fpath: str, binary_id: int = 0) -> Response:
    """
    Get Software Bill Of Materials for binary
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    end_point = f"sboms/{bid}"

    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}", end_point)

    res: Response = reveng_req(requests.get, end_point)

    logger.info("SBOM for %s:\n%s", fpath, res.text)

    res.raise_for_status()
    return res


def RE_binary_additonal_details(fpath: str, binary_id: int = None) -> Response:
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id is None else binary_id
    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}")

    endpoint = f"v2/binaries/{bid}/additional-details"
    res: Response = reveng_req(requests.get, endpoint)
    res.raise_for_status()

    logger.info(f"Additional Details Info({fpath}):\n")
    logger.info(f"\n{json.dumps(res.json(), indent=4)}")
    return res


def RE_binary_details(fpath: str, binary_id: int = None) -> Response:
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id is None else binary_id
    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}")

    endpoint = f"v2/binaries/{bid}/details"
    res: Response = reveng_req(requests.get, endpoint)
    res.raise_for_status()

    logger.info(f"Details Info({fpath}):\n")
    logger.info(f"\n{json.dumps(res.json(), indent=4)}")
    return res


def RE_functions_rename(function_id: int, new_name: str) -> Response:
    """
    Send the new name of a function to C2
    :param function_id: ID of a function
    :param new_name: New function name
    """
    res: Response = reveng_req(
        requests.post,
        f"v1/functions/rename/{function_id}",
        json_data={"new_name": new_name},
    )

    if res.ok:
        logger.info("FunctionId %d has been renamed with '%s'.",
                    function_id, new_name)
    else:
        logger.warning("Error, cannot rename FunctionId %d. %s",
                       function_id, res.text)

    res.raise_for_status()
    return res


def RE_functions_rename_batch(mapping: dict[int, str]) -> Response:
    """
    Send a list of dictionaries, with a corresponding key as function ID and the desired function_name
    :param mapping: dictionary containing the function_id as key and function_name as value
    """
    params = {
        "new_name_mapping": [
            {
                "function_id": func_id,
                "function_name": func_name,
            }
            for func_id, func_name in mapping.items()
        ]
    }

    res: Response = reveng_req(
        requests.post, "v1/functions/batch/rename", json_data=params
    )

    res.raise_for_status()
    return res


def RE_settings() -> Response:
    """
    Get the configuration settings
    """
    res: Response = reveng_req(requests.get, "v1/config")

    res.raise_for_status()
    return res


def RE_health() -> bool:
    """
    Health check & verify access to the API
    """
    res: Response = reveng_req(requests.get, "v1")

    success = res.json()["success"]

    if success:
        logger.info(res.json()["message"])
    else:
        logger.warning(res.json()["error"])
    return success


def RE_authentication() -> Response:
    """
    Authentication Check
    """
    res: Response = reveng_req(requests.get, "v1/authenticate")

    res.raise_for_status()
    return res


def RE_functions_list(
    analysis_id: int,
    search_term: str = "",
    min_v_address: int = 0,
    max_v_address: int = 0,
) -> Response:
    """
    Get the functions of a binary
    :param binary_id: Binary ID
    """
    params = {}
    if search_term:
        params["search_term"] = search_term

    if min_v_address != 0:
        params["min_v_address"] = min_v_address

    if max_v_address != 0:
        params["max_v_address"] = max_v_address

    res: Response = reveng_req(
        requests.get,
        f"/v2/analyses/{analysis_id}/functions/list",
        params=params
    )

    res.raise_for_status()

    return res


def RE_function_callers_callees(function: int) -> Response:
    """
    Get the callers and callees of a functions
    :param function: Function ID
    """
    res: Response = reveng_req(
        requests.get, f"v2/functions/{function}/callees_callers")

    res.raise_for_status()
    return res


def RE_analysis_info(analysis_id: int) -> Response:
    """
    Get the analysis information
    :param analysis_id: Analysis ID
    """
    res: Response = reveng_req(
        requests.get, f"v2/analyses/{analysis_id}/info/basic")

    res.raise_for_status()
    return res


def re_binary_id(fpath: str) -> str:
    """
    Take the SHA-256 hash of binary file
    :param fpath: File path for binary to analyse
    """
    if fpath and isfile(fpath) and access(fpath, R_OK):
        hf = sha256()

        with open(fpath, "rb") as fd:
            c = fd.read()
            hf.update(c)

        return hf.hexdigest()
    else:
        return fpath


def _binary_isa(binary: Binary, exec_type: str) -> str:
    """
    Get ISA format
    """
    if exec_type == "ELF":
        arch = binary.header.machine_type
        if arch == ELF.ARCH.I386:
            return "x86"
        elif arch == ELF.ARCH.X86_64:
            return "x86_64"
        elif arch == ELF.ARCH.ARM:
            return "ARM32"
        elif arch == ELF.ARCH.AARCH64:
            return "ARM64"
    elif exec_type == "PE":
        machine_type = binary.header.machine
        if machine_type == PE.Header.MACHINE_TYPES.I386:
            return "x86"
        elif machine_type == PE.Header.MACHINE_TYPES.AMD64:
            return "x86_64"
        elif machine_type == PE.Header.MACHINE_TYPES.ARM:
            return "ARM32"
        elif machine_type == PE.Header.MACHINE_TYPES.ARM64:
            return "ARM64"
    elif exec_type == "Mach-O":
        cpu_type = binary.header.cpu_type

        if cpu_type == MachO.Header.CPU_TYPE.X86:
            return "x86"
        elif cpu_type == MachO.Header.CPU_TYPE.X86_64:
            return "x86_64"
        elif cpu_type == MachO.Header.CPU_TYPE.ARM:
            return "ARM32"
        elif cpu_type == MachO.Header.CPU_TYPE.ARM64:
            return "ARM64"

    logger.error(
        "Error, could not determine or unsupported "
        f"ISA for binary format: {exec_type}."
    )
    raise RuntimeError(
        "Error, could not determine or unsupported "
        f"ISA for binary format: {exec_type}."
    )


def _binary_format(binary: Binary) -> str:
    """
    Get executable file format
    """
    if binary.format == Binary.FORMATS.PE:
        return "PE"
    if binary.format == Binary.FORMATS.ELF:
        return "ELF"
    if binary.format == Binary.FORMATS.MACHO:
        return "Mach-O"

    logger.error(
        "Error, could not determine or unsupported"
        f" binary format: {binary.format}."
    )
    raise RuntimeError(
        "Error, could not determine or "
        f"unsupported binary format: {binary.format}"
    )


def file_type(fpath: str) -> tuple[str, str]:
    """
    Determine ISA for binary
    :param fpath: File path for binary to analyse
    """
    binary = parse(fpath)

    if not binary:
        file_format = isa_format = "Unknown format"
    else:
        # handle PE and ELF files
        file_format = _binary_format(binary)
        isa_format = _binary_isa(binary, file_format)

    return file_format, isa_format


def parse_config() -> None:
    """
    Parse ~/.reait.toml config file
    """
    fpath = expanduser("~/.reait.toml")

    if isfile(fpath) and access(fpath, R_OK):
        with open(fpath) as fd:
            config = tomli.loads(fd.read())

            for key in (
                "apikey",
                "host",
                "model",
            ):
                if key in config:
                    re_conf[key] = config[key]
    else:
        logger.info("File %s doesn't exist or isn't readable", fpath)


def RE_analysis_id(fpath: str, binary_id: int = 0) -> Response:
    """
    Get the Analysis ID for the Binary ID
    :param fpath: File path for binary to analyse
    :param binary_id: ID of binary
    """
    bin_id = re_binary_id(fpath)
    bid = re_bid_search(bin_id) if binary_id == 0 else binary_id

    end_point = f"v2/analyses/lookup/{bid}"

    if bid == -1:
        raise ReaitError(f"No matches found for hash: {bin_id}", end_point)

    res: Response = reveng_req(requests.get, end_point)

    logger.info("Analysis ID for %s:\n%s", fpath, res.text)

    res.raise_for_status()
    return res


def RE_functions_data_types(
    function_ids: list[int],
) -> Response:
    """
    Get data types for the functions
    :param functions_ids: List of function IDs
    :return: Response object
    """
    endpoint = "/v2/functions/data_types"
    res: Response = reveng_req(
        requests.post, endpoint, json_data={"function_ids": function_ids}
    )
    res.raise_for_status()
    return res


def RE_functions_data_types_poll(
    function_ids: list[int],
) -> Response:
    """
    Poll data types for the functions
    :param functions_ids: List of function IDs
    :return: Response object
    """
    endpoint = "/v2/functions/data_types"
    res: Response = reveng_req(
        requests.get, endpoint, params={"function_ids": function_ids}
    )
    res.raise_for_status()
    return res


def RE_generate_data_types(
    analysis_id: int,
    function_ids: list[int]
) -> Response:
    """
    Generate data types for the analysis
    :param aid: Analysis ID
    """
    end_point = f"/v2/analyses/{analysis_id}/functions/data_types"

    res: Response = reveng_req(
        requests.post, end_point, json_data={"function_ids": function_ids}
    )
    res.raise_for_status()
    return res


def RE_poll_data_types(
    analysis_id: int,
    function_id: int,
) -> Response:
    """
    Poll data types for the analysis
    :param aid: Analysis ID
    """
    end_point = f"/v2/analyses/{analysis_id}/functions/{function_id}/data_types"

    res: Response = reveng_req(requests.get, end_point)
    res.raise_for_status()
    return res


def RE_list_data_types(analysis_id: int, function_ids: list[int]) -> Response:
    """
    List data types for the analysis
    :param aid: Analysis ID
    :param function_ids: List of function IDs
    """
    end_point = f"/v2/analyses/{analysis_id}/functions/data_types"

    res: Response = reveng_req(
        requests.get, end_point, json_data={"function_ids": function_ids}
    )
    res.raise_for_status()
    return res


def RE_begin_ai_decompilation(function_id: int) -> Response:
    """
    Begin AI decompilation for the function
    :param function_id: Function ID
    """
    end_point = f"/v2/functions/{function_id}/ai-decompilation"

    res: Response = reveng_req(
        requests.post,
        end_point,
        data=None,
    )
    res.raise_for_status()
    return res


def RE_poll_ai_decompilation(
    function_id: int,
    summarise: bool = False
) -> Response:
    """
    Poll AI decompilation for the function
    :param function_id: Function ID
    """
    end_point = f"/v2/functions/{function_id}/ai-decompilation"

    params = {}

    if summarise:
        params["summarise"] = summarise

    res: Response = reveng_req(
        requests.get,
        end_point,
        params=params,
    )
    res.raise_for_status()
    return res


def RE_analysis_lookup(binary_id: int) -> Response:
    """
    Get the Analysis ID from a Binary ID
    :param binary_id: Binary ID
    """
    end_point = f"/v2/analyses/lookup/{binary_id}"
    res: Response = reveng_req(requests.get, end_point)
    res.raise_for_status()
    return res


def RE_collections_search(
        page: int = 1,
        page_size: int = 10,
        query: dict = {},
) -> Response:
    """
    Search for collections in the database
    """
    end_point = "/v2/search/collections"
    params = {
        "page": page,
        "page_size": page_size,
    }

    # this api support:
    #   partial_collection_name
    #   partial_binary_name
    #   partial_binary_sha256
    #   model_name
    #   tags

    def exist_and_populated(key: str) -> bool:
        return key in query and query[key] is not None and len(query[key]) > 0

    if exist_and_populated("collection_name"):
        params["partial_collection_name"] = query["collection_name"]
    elif exist_and_populated("binary_name"):
        params["partial_binary_name"] = query["binary_name"]
    elif exist_and_populated("sha_256_hash"):
        params["partial_binary_sha256"] = query["sha_256_hash"]
    elif exist_and_populated("tags"):
        params["tags"] = query["tags"]
    elif exist_and_populated("model_name"):
        params["model_name"] = query["model_name"]
    elif exist_and_populated("query"):
        params["partial_collection_name"] = query["query"]

    res: Response = reveng_req(requests.get, end_point, params=params)
    res.raise_for_status()
    return res


def RE_binaries_search(
        page: int = 1,
        page_size: int = 10,
        query: dict = {},
) -> Response:
    """
    Search for binaries in the database
    """
    end_point = "/v2/search/binaries"
    params = {
        "page": page,
        "page_size": page_size,
    }

    # this api support:
    #   partial_name
    #   partial_sha256
    #   tags
    #   model_name

    def exist_and_populated(key: str) -> bool:
        return key in query and query[key] is not None and len(query[key]) > 0

    if exist_and_populated("binary_name"):
        params["partial_name"] = query["binary_name"]
    elif exist_and_populated("sha_256_hash"):
        params["partial_sha256"] = query["sha_256_hash"]
    elif exist_and_populated("tags"):
        params["tags"] = query["tags"]
    elif exist_and_populated("model_name"):
        params["model_name"] = query["model_name"]
    elif exist_and_populated("query"):
        params["partial_name"] = query["query"]

    res: Response = reveng_req(requests.get, end_point, params=params)
    res.raise_for_status()
    return res


# Bin_id is referred to as hash in this program - to maintain usage BID = id
# of a binary bin_id = hash
# Assumes a file has been passed, correct hash only
# Returns the BID of the binary_id (hash)
def RE_latest_bid(bin_id: str) -> int:
    res: Response = reveng_req(
        requests.get, "v1/search", json_data={"sha_256_hash": bin_id}
    )

    bid = -1

    if res.ok:
        # Filter the result who matches the SHA-256
        binaries = list(
            filter(
                lambda binary: binary["sha_256_hash"] == bin_id,
                res.json()["query_results"],
            )
        )

        # Check only one record is returned
        if len(binaries) == 1:
            binary = binaries[0]
            bid = binary["binary_id"]

            logger.info(
                "Only one record exists, selecting - ID: %d, Name: %s, Creation: %s, Model: %s, Status: %s",
                bid,
                binary["binary_name"],
                binary["creation"],
                binary["model_name"],
                binary["status"],
            )
        elif len(binaries) > 1:
            binaries.sort(
                key=lambda binary: datetime.fromisoformat(
                    binary["creation"]
                ).timestamp(),
                reverse=True,
            )

            logger.info("%d matches found for hash: %s", len(binaries), bin_id)

            options_dict = {}

            for idx, binary in enumerate(binaries):
                logger.info(
                    "[%d] - ID: %d, Name: %s, Creation: %s, Model: %s, Status: %s",
                    idx,
                    binary["binary_id"],
                    binary["binary_name"],
                    binary["creation"],
                    binary["model_name"],
                    binary["status"],
                )

                options_dict[idx] = binary["binary_id"]
            try:
                bid = options_dict[0]
            except Exception:
                bid = options_dict[0]
                logger.warning("Select the most recent analysis - ID: %d", bid)
        else:
            logger.warning("No matches found for hash: %s", bin_id)
    else:
        logger.warning("Bad Request: %s", res.text)

    res.raise_for_status()
    return bid


# NOTE: newest API as per documentation still using /v1/ prefix
def RE_models() -> Response:
    res: Response = reveng_req(requests.get, "v1/models")

    res.raise_for_status()
    return res


# NOTE: newest API as per documentation still using /v1/ prefix
def RE_functions_dump(function_ids: list[int]) -> Response:
    res: Response = reveng_req(
        requests.post, "v1/functions/dump", json_data={"function_id_list": function_ids}
    )

    res.raise_for_status()
    return res


# NOTE: this API endpoint does not actually exist
def RE_generate_summaries(function_id: int) -> Response:
    res: Response = reveng_req(
        requests.get, f"v1/functions/blocks_comments/{function_id}"
    )

    res.raise_for_status()
    return res


def RE_collection_search(search: str) -> Response:
    res: Response = reveng_req(
        requests.get,
        "v1/collections/quick/search",
        params={"search_term": search if search else ""},
    )

    res.raise_for_status()
    return res


def RE_recent_analysis(
    status: str = "All", scope: str = "ALL", nb_analysis: int = 50
) -> Response:
    res: Response = reveng_req(
        requests.get,
        "v1/analyse/recent",
        json_data={"status": status, "scope": scope, "n": nb_analysis},
    )

    res.raise_for_status()
    return res


def RE_search(fpath: str) -> Response:
    bin_id = re_binary_id(fpath)

    res: Response = reveng_req(
        requests.get, "v1/search", json_data={"sha_256_hash": bin_id}
    )

    res.raise_for_status()
    return res


# NOTE: this uses a newer API version
def RE_similar_functions(
    function_id: int,
    limit: int = 20,
    distance: int | float = 0.09999999999999998,
    debug: bool = False,
):
    params = {
        "distance": distance,
        "limit": limit,
        "debug": debug,
    }

    res: Response = reveng_req(
        requests.get, f"v2/functions/{function_id}/similar-functions", params=params
    )

    res.raise_for_status()
    return res



#!/usr/bin/env python3
from __future__ import print_function

from sys import exit, stdout, stderr

import argparse
import json
import logging
import numpy as np
import os
from glob import iglob
from pathlib import Path
from requests import HTTPError
from rich import print_json
from rich.console import Console
from rich.progress import track
from scipy.spatial import distance
from typing import Optional
from reait import api

rerr = Console(file=stderr, width=180)
rout = Console(file=stdout, width=180)


def version() -> int:
    """
    Display program version
    """
    rout.print(
        f"""[bold blue] ::::::::    ::::::::
::  ::::    :::  :::
::::::::::::::::::::
:::::   :::   ::::::
   ::::::::::::::   
   .::  :::  ::::   
::::::  :::  :::::::
::  :::::::::::  :::
::  :::::  ::::  :::
::::::::    :::::::: [/bold blue]
  [bold red]reait[/bold red] [bold bright_green]v{api.__version__}[/bold bright_green]
"""
    )
    rout.print("[yellow]Config:[/yellow]")
    print_json(data=api.re_conf)
    return 0


def verify_binary(fpath_fmt: str) -> tuple[str, str, str]:
    fpath = fpath_fmt

    exec_format, exec_isa = api.file_type(fpath)

    return fpath, exec_format, exec_isa


def match(
    fpath: str, embeddings: list, confidence: float = 0.95, deviation: float = 0.1
) -> None:
    """
    Match embeddings in fpath from a list of embeddings
    """
    rout.print(f"Matching symbols from {fpath} with confidence {confidence}.")
    sink_embed_mat = np.vstack(list(map(lambda x: x["embedding"], embeddings)))
    b_embeds = api.RE_embeddings(fpath).json()["data"]
    source_embed_mat = np.vstack(list(map(lambda x: x["embedding"], b_embeds)))
    # angular distance over cosine
    # closest = 1.0 - distance.cdist(source_embed_mat, sink_embed_mat, 'cosine')
    closest = distance.cdist(source_embed_mat, sink_embed_mat, api.angular_distance)
    # rescale to separate high end of (-1, 1.0)
    # closest = rescale_sim(closest)
    i, j = closest.shape

    for _i in track(range(i), description="Matching Symbols..."):
        row = closest[_i, :]
        match_index, second_match = row.argsort()[::-1][:2]
        source_index = _i
        sink_index = match_index
        source_symb = b_embeds[_i]
        sink_symb = embeddings[sink_index]
        m_confidence = row[match_index]
        s_confidence = row[second_match]

        if row[match_index] >= confidence:
            rout.print(
                f"[bold green]Found match![/bold green][yellow]\tConfidence: {m_confidence:.05f}[/yellow]\t"
                f"[blue]{source_symb['name']}:{source_symb['vaddr']}[/blue]\t->\t"
                f"[blue]{sink_symb['name']}:{sink_symb['vaddr']}"
            )
        elif (m_confidence - s_confidence) > deviation:
            rout.print(
                f"[bold magenta]Possible match[/bold magenta][yellow]\t"
                f"Confidence: {m_confidence:.05f}/{s_confidence:.05f}[/yellow]\t"
                f"[blue]{source_symb['name']}:{source_symb['vaddr']}[/blue]\t->\t"
                f"[blue]{sink_symb['name']}:{sink_symb['vaddr']}"
            )
        else:
            rerr.print(
                f"[bold red]No match for[/bold red]\t[blue]{source_symb['name']}:{source_symb['vaddr']}\t"
                f"{sink_symb['name']} - {m_confidence:0.05f}[/blue]"
            )
            pass


def match_for_each(fpath: str, confidence: float = 0.9, nns: int = 1) -> int:
    """
    Match embeddings in fpath from a list of embeddings
    """
    nns = max(nns, 1)

    rout.print(
        f"Matching symbols from '{fpath}' with a confidence {confidence:.02f} and up to "
        f"{nns} result{'' if nns == 1 else 's'} per function"
    )
    functions = api.RE_analyze_functions(fpath).json()["functions"]
    function_matches = api.RE_nearest_functions(
        fpath, nns=nns, distance=1 - confidence
    ).json()["function_matches"]

    if len(function_matches) == 0:
        rerr.print(
            f"[bold red]No matches found for a confidence of [/bold red] {confidence:.02f}"
        )
        return -1
    else:
        for function in functions:
            matches = list(
                filter(
                    lambda x: function["function_id"] == x["origin_function_id"],
                    function_matches,
                )
            )

            if len(matches):
                rout.print(
                    f"[bold green]Found {len(matches)} match{'' if len(matches) == 1 else 'es'} for "
                    f"[/bold green][blue]{function['function_name']}: {function['function_vaddr']:#x}[/blue]"
                )

                for match in matches:
                    rout.print(
                        f"\t[yellow]Confidence: {match['confidence']:.05f}[/yellow]"
                        f"\t[blue]{match['nearest_neighbor_function_name']}"
                        f" ({match['nearest_neighbor_binary_name']})[/blue]"
                    )
            else:
                rout.print(
                    f"[bold red]No matches found for[/bold red] "
                    f"[blue]{function['function_name']}: {function['function_vaddr']:#x}[/blue]"
                )
    return 0


def parse_collections(collections: str) -> Optional[list[str]]:
    """
    Return collections as list from CSV
    """
    if not collections:
        return None
    return collections.split(",")


def rescale_sim(x):
    """
    Too many values close to 0.999, 0.99999, 0.998, rescale so small values are very low,
    high values separated, map to hyperbolic space
    """
    return np.power(x, 5)


def validate_file(arg):
    file = Path(arg)
    if file.is_file():
        return file.absolute()
    raise FileNotFoundError(f"File path {arg} does not exists.")


def validate_dir(arg):
    dir = Path(arg)
    if dir.is_dir():
        return dir.absolute()
    raise NotADirectoryError(f"Directory path {arg} does not exists.")


def report_api_error_message(f):
    """
    Print message from API errors to console
    """

    def decorate(f):
        def applicator(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except HTTPError as err:
                content = err.response.json()
                if "message" in content:
                    rerr.print(
                        f"[bold red]API Error[/bold red] [bold blue_violet]{err.response.status_code}[/bold blue_violet][bold red] to [/bold red][bold yellow]{err.response.url}[/bold yellow][bold red] {err.response.json()['message']}[/bold red]"
                    )
                    if "errors" in content:
                        for msg in content["errors"]:
                            rerr.print(f"[bold red]{msg['message']}[/bold red]")

                    exit()

        return applicator

    return decorate(f)


@report_api_error_message
def main() -> int:
    """
    Tool entry
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "-b",
        "--binary",
        type=validate_file,
        help="Path of binary to analyse, use ./path:{exec_format} to specify executable format e.g. ./path:raw-x86_64",
    )
    parser.add_argument(
        "-B",
        "--binary-hash",
        default="",
        help="Hex-encoded SHA-256 hash of the binary to use",
    )
    parser.add_argument(
        "-D",
        "--dir",
        type=validate_dir,
        help="Path of directory to recursively analyse",
    )
    parser.add_argument(
        "-a",
        "--analyse",
        action="store_true",
        help="Perform a full analysis and generate embeddings for every symbol",
    )
    parser.add_argument(
        "--base-address",
        help="Image base of the executable image to map for remote analysis",
    )
    parser.add_argument(
        "-A", action="store_true", help="Upload and Analyse a new binary"
    )
    parser.add_argument(
        "-u",
        "--upload",
        action="store_true",
        help="Upload a new binary to remote server",
    )
    parser.add_argument(
        "--duplicate",
        default=False,
        action="store_true",
        help="Duplicate an existing binary",
    )
    parser.add_argument(
        "--details",
        default=False,
        action="store_true",
        help="Get binary additional details",
    )
    parser.add_argument(
        "-e", "--embedding", help="Path of JSON file containing a BinNet embedding"
    )
    parser.add_argument(
        "--nns",
        default="5",
        help="Number of approximate nearest neighbors to fetch",
        type=int,
    )
    parser.add_argument(
        "--collections",
        default=None,
        help="Comma Seperated Value of collections to search from e.g. libxml2,libpcap. Used to select RevEng.AI collections for filtering search results",
    )
    parser.add_argument(
        "--found-in",
        help="ANN flag to limit to embeddings returned to those found in specific binary",
    )
    parser.add_argument(
        "--from-file",
        help="ANN flag to limit to embeddings returned to those found in JSON embeddings file",
    )
    parser.add_argument(
        "-c", "--cves", action="store_true", help="Check for CVEs found inside binary"
    )
    parser.add_argument("--sbom", action="store_true", help="Generate SBOM for binary")
    parser.add_argument(
        "-m", "--model", default=None, help="AI model used to generate embeddings"
    )
    parser.add_argument(
        "-x", "--extract", action="store_true", help="Fetch embeddings for binary"
    )
    parser.add_argument(
        "-M",
        "--match",
        action="store_true",
        help="Match functions in binary file. Can be used with --confidence, --deviation, --from-file, --found-in.",
    )
    parser.add_argument(
        "--confidence",
        default="high",
        choices=["high", "medium", "low", "partial", "all"],
        help="Confidence threshold used to match symbols. Valid values are 'all', 'medium', 'low', 'partial' or 'high'[DEFAULT]",
    )
    parser.add_argument(
        "--deviation",
        default=0.1,
        type=float,
        help="Deviation in prediction confidence between outlier and next highest symbol. Use if confident symbol is present in binary but not matching.",
    )
    parser.add_argument(
        "-l", "--logs", action="store_true", help="Fetch analysis log file for binary"
    )
    parser.add_argument(
        "-d",
        "--delete",
        action="store_true",
        help="Delete all metadata associated with binary",
    )
    parser.add_argument("-k", "--apikey", help="RevEng.AI Personal API key")
    parser.add_argument("-h", "--host", help="Analysis Host (https://api.reveng.ai)")
    parser.add_argument(
        "-v", "--version", action="store_true", help="Display version information"
    )
    parser.add_argument(
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help=argparse._("Show this help message and exit"),
    )
    parser.add_argument(
        "--isa",
        default=None,
        help="Override executable ISA. Valid values are x86, x86_64, ARMv7",
    )
    parser.add_argument(
        "--exec-format",
        default=None,
        help="Override executable format. Valid values are pe, elf, macho, raw",
    )
    parser.add_argument(
        "--platform",
        default=None,
        help="Override OS platform. Valid values are Windows, Linux, OSX, OpenBSD",
    )
    parser.add_argument(
        "--dynamic-execution",
        default=False,
        action="store_true",
        help="Enable dynamic execution in sandbox during analysis. Analysis will include any auto unpacked malware samples",
    )
    parser.add_argument(
        "--cmd-line-args",
        default="",
        help="Command line arguments to pass when running binary sample in the sandbox. Only used when run with --dynamic-execution",
    )
    parser.add_argument(
        "--scope",
        default="private",
        choices=["public", "private"],
        help="Override analysis visibility (scope). Valid values are 'public' or 'private'[DEFAULT]",
    )
    parser.add_argument(
        "--tags",
        default=None,
        type=str,
        help="Assign tags to an analysis. Valid responses are tag1,tag2,tag3.",
    )
    parser.add_argument(
        "--do-not-auto-tag",
        default=False,
        action="store_true",
        help="Disable auto-tagging in API views",
    )
    parser.add_argument(
        "--priority", default=0, type=int, help="Add priority to processing queue."
    )
    parser.add_argument(
        "--verbose", default=False, action="store_true", help="Set verbose output."
    )
    parser.add_argument(
        "--debug", default=None, help="Debug file path to write pass with analysis"
    )
    parser.add_argument(
        "-s",
        "--status",
        action="store_true",
        help="Ongoing status of the provided binary",
    )

    args = parser.parse_args()

    # set re_conf args
    for arg in (
        "apikey",
        "host",
        "model",
    ):
        if getattr(args, arg):
            api.re_conf[arg] = getattr(args, arg)

    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.INFO)

    # display version and exit
    if args.version:
        return version()

    # validate length of string tags
    tags = None
    if args.tags:
        tags = parse_collections(args.tags)

    collections = None
    if args.collections:
        collections = parse_collections(args.collections)

    # auto analysis, uploads and starts analysis
    if args.A:
        args.upload = args.analyse = True

    if args.dir:
        files = iglob(os.path.abspath(args.dir) + "/**/*", recursive=True)
        # perform operation on all files inside directory
        for file in track(files, description="Files in directory"):
            if not os.path.isfile(file):
                rerr.print(f"[blue]Skipping non-file:[/blue] {file}")
                continue

            # upload binary
            if args.upload:
                api.RE_upload(file)

            if args.analyse:
                try:
                    fpath, exec_fmt, exec_isa = verify_binary(file)
                    rout.print(f"Found {fpath}: {exec_fmt}-{exec_isa}")
                except Exception as e:
                    rerr.print(
                        f"[red bold][!] Error, binary exec type could not be verified:[/red bold] {file}"
                    )
                    rerr.print(f"[yellow] {e} [/yellow]")

                rout.print(f"[green bold]Analysing:[/green bold] {file}")
                api.RE_analyse(
                    file,
                    model_name=api.re_conf["model"],
                    isa_options=args.isa,
                    platform_options=args.platform,
                    dynamic_execution=args.dynamic_execution,
                    command_line_args=args.cmd_line_args,
                    file_options=args.exec_format,
                    binary_scope=args.scope.upper(),
                    tags=tags,
                    priority=args.priority,
                    duplicate=args.duplicate,
                    debug_fpath=args.debug,
                    skip_scraping=args.do_not_auto_tag,
                )

            if args.delete:
                try:
                    rout.print(
                        f"[green bold]Deleting analyses for:[/green bold] {file}"
                    )
                    api.RE_delete(file)
                except Exception as e:
                    rerr.print(
                        f"[red bold][!] Error, could not delete analysis for:[/red bold] {file}"
                    )
                    rerr.print(f"[yellow] {e} [/yellow]")

            if not (args.upload or args.analyse or args.delete):
                rerr.print(
                    f"Error, '-D' flag only supports upload, analyse, or delete."
                )
                return -1
    elif (
        args.analyse
        or args.extract
        or args.logs
        or args.delete
        or args.details
        or args.upload
        or args.match
        or args.cves
        or args.sbom
        or args.status
    ):
        try:
            fpath, exec_fmt, exec_isa = verify_binary(args.binary)
            # keep stdout to data only
            rout.print(f"Found {fpath}: {exec_fmt}-{exec_isa}")
            args.binary = fpath
        except TypeError as e:
            rerr.print(
                "[bold red][!] Error, please supply a valid binary file using '-b' flag.[/bold red]"
            )
            rerr.print(f"[yellow] {e} [/yellow]")
            return 0
        except Exception as e:
            rerr.print(
                f"[bold red][!] Error, binary exec type could not be verified:[/bold red] {args.binary}"
            )
            rerr.print(f"[yellow] {e} [/yellow]")

        if args.upload:
            api.RE_upload(args.binary)

            if not args.analyse:
                return 0

        # upload binary first, them carry out actions
        if args.analyse:
            api.RE_analyse(
                args.binary,
                model_name=api.re_conf["model"],
                isa_options=args.isa,
                platform_options=args.platform,
                dynamic_execution=args.dynamic_execution,
                command_line_args=args.cmd_line_args,
                file_options=args.exec_format,
                binary_scope=args.scope.upper(),
                tags=tags,
                priority=args.priority,
                duplicate=args.duplicate,
                debug_fpath=args.debug,
                skip_scraping=args.do_not_auto_tag,
            )

        elif args.extract:
            embeddings = api.RE_embeddings(args.binary).json()
            print_json(data=embeddings)

        elif args.match:
            # parse confidences
            confidence: float = 0.90
            if args.confidence:
                confidences = {
                    "high": 0.95,
                    "medium": 0.9,
                    "low": 0.7,
                    "partial": 0.5,
                    "all": 0.0,
                }
                if args.confidence in confidences.keys():
                    confidence = confidences[args.confidence]

            if args.from_file:
                if not os.path.isfile(args.from_file) and not os.access(
                    args.from_file, os.R_OK
                ):
                    rerr.print(
                        "[bold red][!] Error, '--from-file' flag requires a path to a JSON embeddings file.[/bold red]"
                    )
                    return -1
                rout.print(
                    f"[+] Searching for symbols similar to embedding in binary: {args.from_file}"
                )
                embeddings = json.load(open(args.from_file))
            elif args.found_in:
                if not os.path.isfile(args.found_in) and not os.access(
                    args.found_in, os.R_OK
                ):
                    rerr.print(
                        "[bold red][!] Error, '--found-in' flag requires a path to a binary to search from.[/bold red]"
                    )
                    return -1
                rout.print(
                    f"[+] Matching symbols between {args.binary} and {args.found_in}."
                )
                embeddings = api.RE_embeddings(args.found_in).json()["data"][
                    "embedding"
                ]
            else:
                return match_for_each(args.binary, confidence, args.nns)

            match(
                args.binary,
                embeddings,
                confidence=confidence,
                deviation=float(args.deviation),
            )

        elif args.logs:
            api.RE_logs(args.binary)

        elif args.delete:
            api.RE_delete(args.binary)

        elif args.sbom:
            api.RE_SBOM(args.binary)

        elif args.cves:
            api.RE_cves(args.binary)
        elif args.details:
            api.RE_binary_additonal_details(args.binary)

        elif args.status:
            api.RE_status(args.binary, console=True)

    else:
        rerr.print("[bold red][!] Error, please supply an action command.[/bold red]")
        parser.print_help()
    return 0


if __name__ == "__main__":
    exit(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from subprocess import check_call
from sys import executable
from unittest import main

from utils import BaseTestCase, testlog

import reait.api as api


def run_test_script(fpath: str, *args) -> int:
    cmd = [executable, fpath] + list(args)

    testlog.info("Running '%s'", " ".join(cmd))
    return check_call(cmd, timeout=60)


class TestReait(BaseTestCase):
    def test_1_version(self):
        try:
            self.assertEqual(0,
                             run_test_script("src/reait/main.py", "--version"))
        except Exception as e:
            testlog.error("Something went wrong when displaying version. %s", e)

    def test_2_upload(self):
        try:
            self.assertEqual(0,
                             run_test_script("src/reait/main.py",
                                             "--binary", self._fpath, "--upload",
                                             "--apikey", api.re_conf["apikey"]))
        except Exception as e:
            testlog.error("Something went wrong when upload binary for analysis. %s", e)

    def test_3_analyse(self):
        try:
            self.assertEqual(0,
                             run_test_script("src/reait/main.py",
                                             "--binary", self._fpath, "--analyse", "--duplicate",
                                             "--apikey", api.re_conf["apikey"],
                                             "--model", api.re_conf["model"]))
        except Exception as e:
            testlog.error("Something went wrong when start analysis. %s", e)
        finally:
            self._cleanup_binaries(self._fpath)

    def test_4_upload_analyse(self):
        try:
            self.assertEqual(0,
                             run_test_script("src/reait/main.py",
                                             "--binary", self._fpath, "-A", "--duplicate",
                                             "--apikey", api.re_conf["apikey"],
                                             "--model", api.re_conf["model"]))
        except Exception as e:
            testlog.error("Something went wrong when upload + start analysis. %s", e)

    def test_5_logs(self):
        try:
            self.assertEqual(0,
                             run_test_script("src/reait/main.py",
                                             "--binary", self._fpath, "--logs",
                                             "--apikey", api.re_conf["apikey"]))
        except Exception as e:
            testlog.error("Something went wrong when getting logs analysis. %s", e)

    def test_6_status(self):
        try:
            self.assertEqual(0,
                             run_test_script("src/reait/main.py",
                                             "--binary", self._fpath, "--status",
                                             "--apikey", api.re_conf["apikey"]))
        except Exception as e:
            testlog.error("Something went wrong when getting status. %s", e)

    def test_7_delete(self):
        try:
            self.assertEqual(0,
                             run_test_script("src/reait/main.py",
                                             "--binary", self._fpath, "--delete",
                                             "--apikey", api.re_conf["apikey"]))
        except Exception as e:
            testlog.error("Something went wrong when deleting analysis. %s", e)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from sys import path
from pathlib import Path
from unittest import main

# Make it possible to run this file from the root dir of reait without installing reait
path.insert(0, Path(__file__).parent.as_posix())

from utils import BaseTestCase

import reait.api as api


class TestAPIs(BaseTestCase):
    def test_0_conf(self):
        self.assertGreaterEqual(len(api.re_conf), 3)
        self.assertTrue(all(api.re_conf[key] for key in ("apikey", "host", "model",)))
        self.assertNotEqual(api.re_conf["apikey"], "l1br3")
        self.assertIn(self._platform, api.re_conf["model"])

    def test_1_upload(self):
        try:
            response = api.RE_upload(self._fpath).json()

            self.assertTrue(response["success"], "Upload file has failed")
            self.assertEqual(response["sha_256_hash"], api.re_binary_id(self._fpath), "SHA-256 mismatch")
        except Exception:
            self.fail(f"Failed to upload {self._fpath}")

    def test_2_analysis(self):
        try:
            response = api.RE_analyse(self._fpath, model_name=api.re_conf["model"], duplicate=True).json()

            self.assertTrue(response["success"], "Analysis file has failed")
            self.assertIsInstance(response["binary_id"], int)
        except Exception:
            self.fail(f"Failed to analyse {self._fpath}")

    def test_3_analysis_failure(self):
        try:
            # Should raise a ReaitError because of duplicate analysis
            api.RE_analyse(self._fpath, model_name=api.re_conf["model"])

            self.fail(f"Duplicate analysis for {self._fpath}")
        except Exception as e:
            self.assertIsInstance(e, api.ReaitError)
            self.assertIsNotNone(e.response)
            self.assertEqual(e.response.status_code, 404)
            self.assertFalse(e.response.json()["success"])

    def test_4_logs(self):
        try:
            response = api.RE_logs(self._fpath).json()

            self.assertTrue(response["success"], "Analysis file has failed")
            self.assertIsNotNone(response["logs"], "Empty logs analysis")
        except Exception:
            self.fail("Failed to retrieve logs")

    def test_5_delete(self):
        try:
            response = api.RE_delete(self._fpath).json()

            self.assertTrue(response["success"], "Delete file has failed")
        except Exception:
            self.fail(f"Failed to delete {self._fpath}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from setuptools import setup, find_packages


with open("requirements.txt") as fd:
    required = fd.read().splitlines()

with open("README.md", encoding="utf-8") as fd:
    long_description = fd.read()


setup(
    name="reait",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RevEng-AI/reait",
    author="James Patrick-Evans",
    author_email="james@reveng.ai",
    platforms="Cross Platform",
    packages=find_packages(
        where="src",
        exclude=[
            "tests",
        ],
    ),
    package_dir={
        "": "src",
    },
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    install_requires=required,
)
